# pylint: disable=invalid-name
"""Utilities package for my other projects."""
from __future__ import annotations

name = "wg_utilities"

__version__ = "3.8.0"

__all__ = [
    "api",
    "clients",
    "devices",
    "exceptions",
    "functions",
    "loggers",
    "testing",
]
